use crate::prepare_pattern::prepare_pattern;
use libc;
use std::fs::File;
use std::io::{BufReader, Bytes};
use std::path::Path;

const PARITY: u8 = 0o200;
const RECORD_SEPARATOR: u8 = 30;
const OFFSET: u8 = 14;
const BIGRAMS: usize = BIGRAMS;

#[derive(Clone, Debug)]
pub struct Squozen {
    bigrams: [char; BIGRAMS],
    path: Path,
}

impl Squozen {
    pub fn new(filename: Path) -> Result<Squozen, Error> {
        let mut dbfile = File::open(filename)?;
        let mut db = BufReader::<File>::new(dbfile);
        let mut bigrams: [char; BIGRAMS] = [0; BIGRAMS];
        db.read_exact(&mut bigrams)?;
        Ok(Squozen {
            bigrams,
            path: filename.clone(),
        })
    }

    pub fn paths(&mut self) -> StoredPath {
        StoredPath::new(&self);
    }

    /*
    pub fn matches(&mut self, pattern: &[u8]) -> StoredPath {
        FoundPath::new(&self, pattern);
    }
    */
}

pub struct StoredPath<'a> {
    source: &'a Squozen,
    path: [char; libc::PATH_MAX],
    db: Bytes<BufReader<File>>,
    ch: u8,
    last: usize,
    found: bool,
}

impl<'a> StoredPath<'a> {
    pub fn new(squozen: &'a mut Squozen) -> StoredPath {
        let mut dbfile = File::open(&squozen.path)?;
        let mut dbbuffer = BufReader::<File>::new(dbfile);
        dbbuffer.seek(BIGRAMS);
        let mut db = dbbuffer.bytes();
        let mut ch = db.next()?;
        StoredPath {
            squozen,
            db,
            ch,
            path: [0; libc::PATH_MAX],
            last: 0,
        }
    }

    // Note that in either case, the file pointer will be pointing at the first
    // valid character of the database, or it will be over.
    pub fn getw(&mut self) -> Result<u16> {
        let ch1 = self.db.next()?;
        let ch2 = self.db.next()?;
        Ok(u16::from_le_bytes(&[ch1, ch2]))
    }

    pub fn get_offset(&mut self) -> Result<usize> {
        if self.ch == RECORD_SEPARATOR {
            let offset = self.getw()?;
            Ok(usize::from(offset))
        } else {
            Ok(usize::from(self.ch))
        }
    }
}

impl<'a> Iterator for StoredPath<'a> {
    type Item = &'a [char; libc::PATH_MAX];

    fn next(&mut self) -> Option<&Self::Item> {
        let offset = self.get_offset();
        let position = 0;
        loop {
            self.ch = self.db.next()?;
            if self.ch <= RECORD_SEPARATOR {
                break;
            }
            if self.ch < PARITY {
                self.path[self.last + position] = self.ch;
                position += 1;
            } else {
                let bg = self.ch & PARITY - 1;
                self.path[self.last + position] = self.squozen.bigrams[bg * 2];
                self.path[self.last + position + 1] = self.squozen.bigrams[bg * 2 + 1];
                position += 2;
            }
        }
        Some(&self.path)
    }
}

pub struct FoundPath<'a> {
    stored: StoredPath<'a>,
    pattern: Vec<u8>,
}

impl<'a> FoundPath<'a> {
    pub fn new(squozen: &'a Squozen, path: &[u8]) -> Result<FoundPath<'a>, Error> {
        Ok(FoundPath {
            stored: StoredPath::new(squozen),
            pattern: prepare_pattern(path)?,
        })
    }
}

impl<'a> Iterator for FoundPath<'a> {
    type Item = &'a [char; libc::PATH_MAX];

    fn next(&mut self) -> Option<&Self::Item> {
        self.stored.next()
    }
}
